(custom-set-faces
;; custom-set-faces was added by Custom.
;; If you edit it by hand, you could mess it up, so be careful.
;; Your init file should contain only one such instance.
;; If there is more than one, they won't work right.
'(default ((t (:stipple nil :background "black" :foreground "Wheat" :inverse-video nil :box nil :strike-through nil :overline nil :underline nil :slant normal :weight regular :height 158 :width normal :family "Inconsolata" :foundry "CYRE"))))
'(background "blue")
'(border ((t (:background "grey60"))))
'(cursor ((t (:background "white"))))
'(fixed-pitch ((t (:family "Inconsolata, Monaco, Fira Code Retina,LXGW WenKai Hangul" :height 180))))
'(font-lock-builtin-face ((((class color) (background dark)) (:foreground "Turquoise"))))
'(font-lock-comment-delimiter-face ((default (:inherit font-lock-comment-face)) (((class color) (min-colors 16)) nil)))
'(font-lock-comment-face ((t (:foreground "MediumAquamarine"))))
'(font-lock-constant-face ((t (:foreground "green4" :weight bold))))
'(font-lock-doc-face ((t (:inherit font-lock-string-face :foreground "light blue"))))
'(font-lock-doc-string-face ((t (:foreground "green2"))))
'(font-lock-function-name-face ((t (:foreground "gold"))))
'(font-lock-keyword-face ((t (:foreground "royal blue"))))
'(font-lock-negation-char-face ((t (:foreground "grey80"))))
'(font-lock-preprocessor-face ((t (:foreground "light green" :slant normal))))
'(font-lock-reference-face ((t (:foreground "DodgerBlue"))))
'(font-lock-regexp-grouping-backslash ((t (:inherit bold :background "ivory" :foreground "green4"))))
'(font-lock-regexp-grouping-construct ((t (:inherit bold :background "ivory" :foreground "green3"))))
'(font-lock-string-face ((t (:foreground "red"))))
'(font-lock-type-face ((t (:foreground "cyan"))))
'(font-lock-variable-name-face ((t (:foreground "green"))))
'(font-lock-warning-face ((((class color) (background dark)) (:foreground "yellow" :background "red"))))
'(fringe ((((class color) (background dark)) (:background "grey30"))))
'(highlight ((t (:background "CornflowerBlue"))))
'(hl-line ((t (:inherit highlight :background "grey20"))))
'(list-mode-item-selected ((t (:background "gold"))))
'(makefile-space-face ((t (:background "wheat"))) t)
'(menu ((((type x-toolkit)) (:background "black" :foreground "white" :underline "white" :family "Inconsolata"))))
'(mode-line ((t (:background "Navy"))))
'(mode-line-buffer-id ((t (:weight bold))))
'(org-block ((t (:inherit fixed-pitch))))
'(org-code ((t (:inherit (shadow fixed-pitch)))))
'(org-document-info ((t (:foreground "dark orange"))))
'(org-document-info-keyword ((t (:inherit (shadow fixed-pitch)))))
'(org-document-title ((t (:inherit default :weight bold :foreground "Wheat" :family "Inconsolata, LXGW WenKai Hangul,WenQuanYi Micro Hei, Sans Serif" :height 2.0 :underline nil))))
'(org-indent ((t (:inherit (org-hide fixed-pitch)))))
'(org-level-1 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco" :height 1.75))))
'(org-level-2 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco" :height 1.5))))
'(org-level-3 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco" :height 1.25))))
'(org-level-4 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco" :height 1.1))))
'(org-level-5 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco"))))
'(org-level-6 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco"))))
'(org-level-7 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco"))))
'(org-level-8 ((t (:inherit default :weight bold :foreground "Wheat" :family "LXGW WenKai Hangul,WenQuanYi Micro Hei,Inconsolata, monaco"))))
'(org-link ((t (:foreground "royal blue" :underline t))))
'(org-meta-line ((t (:inherit (font-lock-comment-face fixed-pitch)))))
'(org-property-value ((t (:inherit fixed-pitch))))
'(org-special-keyword ((t (:inherit (font-lock-comment-face fixed-pitch)))))
'(org-table ((t (:inherit fixed-pitch :foreground "#83a598"))))
'(org-tag ((t (:inherit (shadow fixed-pitch) :weight bold :height 0.8))))
'(org-verbatim ((t (:inherit (shadow fixed-pitch)))))
'(paren-match ((t (:background "darkseagreen4"))))
'(region ((t (:background "DarkSlateBlue"))))
'(scroll-bar ((t (:background "grey75"))))
'(show-paren-match ((t (:foreground "black" :background "wheat"))))
'(show-paren-mismatch ((((class color)) (:foreground "white" :background "red"))))
'(speedbar-button-face ((((class color) (background dark)) (:foreground "green4"))))
'(speedbar-directory-face ((((class color) (background dark)) (:foreground "khaki"))))
'(speedbar-file-face ((((class color) (background dark)) (:foreground "cyan"))))
'(speedbar-tag-face ((((class color) (background dark)) (:foreground "Springgreen"))))
'(tool-bar ((default (:background "black" :foreground "white")) (((type x w32 mac) (class color)) (:background "grey80"))))
'(variable-pitch ((t (:family "LXGW WenKai Hangul,monaco" :height 180))))
'(vertical-border ((nil nil)))
'(vhdl-speedbar-architecture-selected-face ((((class color) (background dark)) (:underline t :foreground "Blue"))))
'(vhdl-speedbar-entity-face ((((class color) (background dark)) (:foreground "darkGreen"))))
'(vhdl-speedbar-entity-selected-face ((((class color) (background dark)) (:underline t :foreground "darkGreen"))))
'(vhdl-speedbar-package-face ((((class color) (background dark)) (:foreground "black"))))
'(vhdl-speedbar-package-selected-face ((((class color) (background dark)) (:underline t :foreground "black"))))
'(widget-field ((((class grayscale color) (background light)) (:background "DarkBlue"))))
'(widget-inactive ((t (:inherit shadow)))))
